import os
import random
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from collections.abc import Callable
from typing import List
from matplotlib.axes import Axes


def getRandomColor():
    """获取一个随机的颜色"""
    colorArr = ['1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F']
    color = ""
    for i in range(6):
        color += colorArr[random.randint(0, 14)]
    return "#" + color


class NodeItem:
    def __init__(self, nodeName, nodeId, label: str = None):
        self.nodeName = nodeName
        self.nodeId = nodeId
        self.label = nodeName if label is None else label


class GraphBase:
    """
    一个基于 matplotlib 实现的通用绘图接口
    """

    def __init__(self):
        # 记录所有可用的 Marker
        # https://matplotlib.org/stable/api/markers_api.html
        self.markers = ["o", "v", "^", "<", ">", "1", "2", "3", "4", "8", "s", "p", "P", "*", "h", "H", "+",
                        "x", "X", "D", "d", "|", "_", ".", ","]
        # 记录当前图已经使用的Marker
        self.currentUsedMarker = {}
        # 记录常用的 color
        self.colors = [
            "tab:blue", "tab:orange", "tab:green", "tab:red", "tab:purple",
            "tab:brown", "tab:pink", "tab:gray", "tab:olive", "tab:cyan", "red", "green", "blue", "c", "m", "y",
        ]
        # 记录当前图已经使用的 color
        self.currentUsedColor = {}
        # 记录常用的 hatch
        self.hatches = ['', '/', 'o', 'x', '.', '\\', 'O', '|', '-', '+', 'x', '*',
                        '/o', '\\|', '|*', '-\\', '+o', 'x*', 'o-', 'O|', 'O.', '*-']
        self.currentUsedHatch = {}
        # 字体大小
        self.fontSize = 12
        # plotFilter
        self.plotFilter: Callable[[float, float, str, str, str], bool] = None
        # platMap
        self.plotMap: Callable[[List[float], List[float], str, str, str], (List[float], List[float])] = None
        # bindAx
        self.ax: Axes = None

    def bindAxes(self, ax: Axes):
        """
        绑定一个 ax 对象，用于绘图
        :param ax:
        :return:
        """
        self.ax = ax
        return self

    def setPlotFilter(self, plotFilter: Callable[[float, float, str, str, str], bool]):
        """
        设置 plot 函数过滤器，
        [x_value, y_value, nodeName, nodeId, label] => bool value indicate is selected
        :return:
        """
        self.plotFilter = plotFilter
        return self

    def setPlotMap(self, plotMap: Callable[[List[float], List[float], str, str, str], (List[float], List[float])]):
        """
        设置 plot 转换器
        [x, y, nodeName, nodeId, label] => [new_x, new_y]
        :param plotMap:
        :return:
        """
        self.plotMap = plotMap
        return self

    def setFontSize(self, fontSize: int):
        """
        设置字体大小
        :param fontSize:
        :return:
        """
        self.fontSize = fontSize
        return self

    def autoSelectHatch(self) -> str:
        """
        从所有的候选 Hatch 中自动选择
        :return:
        """
        for hatch in self.hatches:
            if hatch not in self.currentUsedHatch:
                self.currentUsedHatch[hatch] = hatch
                return hatch
        return ""

    def autoSelectMarker(self) -> str:
        """
        从所有的marker中自动选择 Marker，保证不和已有的Marker重复
        :return:
        """
        for marker in self.markers:
            if marker not in self.currentUsedMarker:
                self.currentUsedMarker[marker] = marker
                return marker
        return ""

    def autoSelectColor(self) -> str:
        for color in self.colors:
            if color not in self.currentUsedColor:
                return color
        while True:
            randomColor = getRandomColor()
            if randomColor not in self.currentUsedColor:
                break
            self.currentUsedColor[randomColor] = randomColor
        return randomColor

    def tick_direction(self, direction: str):
        """
        设置坐标轴 ticker 的方向，可以取 ['in', 'out', 'inout']
        :param direction:
        :return:
        """
        if self.ax is not None:
            pass
            self.ax.tick_params(direction=direction)
        else:
            plt.rcParams['xtick.direction'] = direction
            plt.rcParams['ytick.direction'] = direction
        return self

    def grid(self, *args, **kwargs):
        if self.ax is not None:
            self.ax.grid(*args, **kwargs)
        else:
            plt.grid(*args, **kwargs)
        return self

    def xlim(self, *args, **kwargs):
        if self.ax is not None:
            self.ax.set_xlim(*args, **kwargs)
        else:
            plt.xlim(*args, **kwargs)
        return self

    def ylim(self, *args, **kwargs):
        if self.ax is not None:
            self.ax.set_ylim(*args, **kwargs)
        else:
            plt.ylim(*args, **kwargs)
        return self

    def ticker_params(self, axis='both', **kwargs):
        if self.ax is not None:
            self.ax.tick_params(axis, **kwargs)
        else:
            plt.tick_params(axis, **kwargs)
        return self

    def xticks(self, ticks=None, labels=None, **kwargs):
        if 'fontsize' not in kwargs:
            kwargs['fontsize'] = self.fontSize
        if self.ax is not None:
            self.ax.tick_params(labelsize=self.fontSize)
            if ticks is not None:
                self.ax.set_xticks(ticks, labels, **kwargs)
        else:
            plt.xticks(ticks, labels, **kwargs)
        return self

    def yticks(self, ticks=None, labels=None, **kwargs):
        if 'fontsize' not in kwargs:
            kwargs['fontsize'] = self.fontSize
        if self.ax is not None:
            self.ax.tick_params(labelsize=self.fontSize)
            if ticks is not None:
                self.ax.set_yticks(ticks, labels, **kwargs)
        else:
            plt.yticks(ticks, labels, **kwargs)
        return self

    def title(self, label, fontdict=None, loc=None, pad=None, **kwargs):
        if 'fontsize' not in kwargs:
            kwargs['fontsize'] = self.fontSize
        if self.ax is not None:
            self.ax.set_title(label, fontdict=fontdict, loc=loc, pad=pad, **kwargs)
        else:
            plt.title(label, fontdict=fontdict, loc=loc, pad=pad, **kwargs)
        return self

    def xlabel(self, xlabel, fontdict=None, labelpad=None, loc=None, **kwargs):
        if 'fontsize' not in kwargs:
            kwargs['fontsize'] = self.fontSize
        if self.ax is not None:
            self.ax.set_xlabel(xlabel, fontdict, labelpad, loc=loc, **kwargs)
        else:
            plt.xlabel(xlabel, fontdict, labelpad, loc=loc, **kwargs)
        return self

    def ylabel(self, ylabel, fontdict=None, labelpad=None, loc=None, **kwargs):
        if 'fontsize' not in kwargs:
            kwargs['fontsize'] = self.fontSize
        if self.ax is not None:
            self.ax.set_ylabel(ylabel, fontdict, labelpad, loc=loc, **kwargs)
        else:
            plt.ylabel(ylabel, fontdict, labelpad, loc=loc, **kwargs)
        return self

    def legend(self, markerSize: int = None, markerEdgeWidth: int = None,
               *args, **kwargs):
        if 'fontsize' not in kwargs:
            kwargs['fontsize'] = self.fontSize
        if self.ax is not None:
            lgnd = self.ax.legend(*args, **kwargs)
        else:
            lgnd = plt.legend(*args, **kwargs)
        if markerSize is not None:
            for leg in lgnd.get_lines():
                leg.set_markersize(markerSize)
        if markerEdgeWidth is not None:
            for leg in lgnd.get_lines():
                leg.set_markeredgewidth(markerEdgeWidth)
        return self

    def _do_filter(self, filter_func, x: List[float], y: List[float], *args):
        cp_x, cp_y = x.copy(), y.copy()
        x, y = [], []
        for i in range(len(cp_x)):
            if filter_func(cp_x[i], cp_y[i], *args):
                x.append(cp_x[i])
                y.append(cp_y[i])
        return x, y

    def _do_map(self, map_func, x: List[float], y: List[float], *args):
        return map_func(x, y, *args)

    def innerPlot(self, x, y, *args,
                  color: str = "&&",
                  linewidth: float = 2, linestyle: str = "dotted", marker: str = "&&",
                  markerfacecolor: str = "none", markersize: float = 6,
                  **kwargs):
        if "linewidth" not in kwargs:
            kwargs["linewidth"] = linewidth
        if "linestyle" not in kwargs:
            kwargs["linestyle"] = linestyle
        # 默认自动选择Marker
        if marker == "&&":
            marker = self.autoSelectMarker()
            self.currentUsedMarker[marker] = marker
        if "marker" not in kwargs:
            kwargs["marker"] = marker
        if "markerfacecolor" not in kwargs:
            kwargs["markerfacecolor"] = 'none'
        if "markersize" not in kwargs:
            kwargs["markersize"] = markersize
        # 默认自动选择 color
        if color == "&&":
            color = self.autoSelectColor()
            self.currentUsedColor[color] = color
        if "color" not in kwargs:
            kwargs["color"] = color

        nodeName, nodeId, label = kwargs.pop("nodeName", ""), kwargs.pop("nodeId", ""), kwargs.get("label", "")
        if self.plotFilter is not None:
            x, y = self._do_filter(self.plotFilter, x, y, nodeName, nodeId, label)

        if self.plotMap is not None:
            x, y = self._do_map(self.plotMap, x, y, nodeName, nodeId, label)

        if self.ax is not None:
            self.ax.plot(x, y, *args, **kwargs)
        else:
            plt.plot(x, y, *args, **kwargs)
        return self

    def innerPlotSum(self, nodeList: [NodeItem], getNode, getX, getY, *args,
                     color: str = "&&",
                     linewidth: float = 2, linestyle: str = "dotted", marker: str = "&&",
                     markerfacecolor: str = "none", markersize: float = 6,
                     **kwargs):
        xMerge = []
        yArr = []
        for nodeItem in nodeList:
            node = getNode(nodeItem.nodeName)
            if not node:
                print("not exist node: ", node)
                continue

            x, y = getX(node, nodeItem.nodeId), \
                   getY(node, nodeItem.nodeId)

            if len(x) > len(xMerge):
                xMerge = x
            yArr.append(np.array(y))

        if len(yArr) <= 0:
            return self

        yMerge = np.zeros(len(xMerge))
        # pad y
        for y in yArr:
            y = np.pad(y, (0, len(xMerge) - len(y)), 'constant', constant_values=(0, 0))
            yMerge += y

        if "label" not in kwargs:
            kwargs["label"] = "Merge"

        self.innerPlot(xMerge, yMerge, *args,
                       color=color,
                       linewidth=linewidth,
                       linestyle=linestyle,
                       marker=marker,
                       markerfacecolor=markerfacecolor,
                       markersize=markersize,
                       **kwargs)
        return self

    def innerPlotAvg(self, nodeList: [NodeItem], getNode, getX, getY, *args,
                     color: str = "&&",
                     linewidth: float = 2, linestyle: str = "dotted", marker: str = "&&",
                     markerfacecolor: str = "none", markersize: float = 6,
                     **kwargs):
        xMerge = []
        yArr = []

        for nodeItem in nodeList:
            node = getNode(nodeItem.nodeName)
            x, y = getX(node, nodeItem.nodeId), \
                   getY(node, nodeItem.nodeId)
            if not node:
                print("not exist node: ", node)
                continue
            if len(x) > len(xMerge):
                xMerge = x
            yArr.append(np.array(y))

        if len(yArr) <= 0:
            return self

        yMerge = np.zeros(len(xMerge))

        # pad y
        for y in yArr:
            y = np.pad(y, (0, len(xMerge) - len(y)), 'constant', constant_values=(0, 0))
            yMerge += y
        yMerge /= len(yArr)

        if "label" not in kwargs:
            kwargs["label"] = "Merge"

        if self.plotFilter is not None:
            x, y = self._do_filter(self.plotFilter, x, y, nodeName, nodeId)

        if self.plotMap is not None:
            x, y = self._do_map(self.plotMap, x, y, nodeName, nodeId)

        self.innerPlot(xMerge, yMerge, *args,
                       color=color,
                       linewidth=linewidth,
                       linestyle=linestyle,
                       marker=marker,
                       markerfacecolor=markerfacecolor,
                       markersize=markersize,
                       **kwargs)
        return self

    def show(self, *args, **kwargs):
        """
        绘制吞吐量图，并show
        :return:
        """
        plt.show(*args, **kwargs)
        return self

    def drawAndSave(self, dir: str, name: str, *args, **kwargs):
        """
        绘制吞吐量图，并保存到图片当中
        :return:
        """
        if not os.path.exists(dir):
            os.makedirs(dir)
        # plt.tight_layout()
        plt.savefig(f"{dir}{os.sep}{name}", *args, **kwargs)
        return self

    def close(self):
        plt.close()
        return self


class ExportItem:
    def __init__(self, nodeName: str,
                 units: str, label: str, x: [float], y: [float]):
        self.nodeName = nodeName
        self.units = units
        self.label = label
        self.x = x
        self.y = y


class ExportBase:
    """
    导出基类，用于将统计数据导出到Excel当中
    """

    def __init__(self, getNode=None, getX=None, getY=None, getUnit=None):
        self.items: [ExportItem] = []
        self.getNode = getNode
        self.getX = getX
        self.getY = getY
        self.getUnit = getUnit

    def select(self, nodeName: str, faceId: any, label: str = None):
        if self.getNode is None or self.getX is None or self.getUnit is None:
            print("getNode is None, can not invoke select")
            return self
        node = self.getNode(nodeName)
        if not node:
            print("not exist node: ", node)
            return self
        if label is None:
            label = nodeName
        x, y = self.getX(node, faceId), self.getY(node, faceId)
        self.items.append(
            ExportItem(
                nodeName,
                self.getUnit(),
                label,
                x,
                y))
        return self

    def selectAvg(self, nodeList: [NodeItem], label: str = None):
        if self.getNode is None or self.getX is None or self.getUnit is None:
            print("getNode is None, can not invoke selectAvg")
            return self
        xMerge = []
        yArr = []

        for nodeItem in nodeList:
            node = self.getNode(nodeItem.nodeName)
            x, y = self.getX(node, nodeItem.nodeId), self.getY(node, nodeItem.nodeId)
            if not node:
                print("not exist node: ", node)
                continue
            if len(x) > len(xMerge):
                xMerge = x
            yArr.append(np.array(y))

        if len(yArr) <= 0:
            return self

        yMerge = np.zeros(len(xMerge))

        # pad y
        for y in yArr:
            y = np.pad(y, (0, len(xMerge) - len(y)), 'constant', constant_values=(0, 0))
            yMerge += y
        yMerge /= len(yArr)
        self.items.append(
            ExportItem(
                label,
                self.getUnit(),
                label,
                xMerge,
                yMerge))
        return self

    def selectSum(self, nodeList: [NodeItem], label: str = None):
        if self.getNode is None or self.getX is None or self.getUnit is None:
            print("getNode is None, can not invoke selectSum")
            return self
        xMerge = []
        yArr = []

        for nodeItem in nodeList:
            node = self.getNode(nodeItem.nodeName)
            x, y = self.getX(node, nodeItem.nodeId), self.getY(node, nodeItem.nodeId)
            if not node:
                print("not exist node: ", node)
                continue
            if len(x) > len(xMerge):
                xMerge = x
            yArr.append(np.array(y))

        if len(yArr) <= 0:
            return self

        yMerge = np.zeros(len(xMerge))

        # pad y
        for y in yArr:
            y = np.pad(y, (0, len(xMerge) - len(y)), 'constant', constant_values=(0, 0))
            yMerge += y
        self.items.append(
            ExportItem(
                label,
                self.getUnit(),
                label,
                xMerge,
                yMerge))
        return self

    def _get_statistical_info_dataframe(self):
        statistics_columns = [
            'Node', 'Label', 'Units', 'Sample Counts',
            'Sum', 'Avg', 'Mean', 'Var', 'std'
        ]
        statistics_data = []
        for item in self.items:
            statistics_data.append([
                item.nodeName,
                item.label,
                item.units,
                len(item.y),
                round(np.sum(item.y), 3),
                round(np.average(item.y), 3),
                round(np.mean(item.y), 3),
                round(np.var(item.y), 3),
                round(np.std(item.y), 3)
            ])

        return pd.DataFrame(statistics_data, columns=statistics_columns)

    def save_statistical_info_as_csv(
            self, save_dir: str, filename: str
    ):
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        self._get_statistical_info_dataframe(). \
            to_csv(f"{save_dir}{os.sep}{filename}")
        return self

    def save_statistical_info_as_excel(
            self, save_dir: str, filename: str
    ):
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        self._get_statistical_info_dataframe(). \
            to_excel(f"{save_dir}{os.sep}{filename}")
        return self

    def print_statistical_info_as_latex(self):
        print(self._get_statistical_info_dataframe().style.to_latex())
        return self

    def print_statistical_info_as_table(self, title: str):
        """
        以表格的形式打印统计信息
        :return:
        """
        from rich.console import Console
        from rich.table import Table
        console = Console()

        table = Table(show_header=True,
                      header_style="bold magenta",
                      title=title, expand=True)
        table.add_column("Node", justify="center", no_wrap=True)
        table.add_column("Label", justify="center", no_wrap=True)
        table.add_column("Units", justify="center", no_wrap=True)
        table.add_column("Sample Counts", justify="center", no_wrap=True)
        table.add_column("Sum", justify="center", no_wrap=True)
        table.add_column("Average", justify="center", no_wrap=True)
        table.add_column("Mean", justify="center", no_wrap=True)
        table.add_column("Var", justify="center", no_wrap=True)
        table.add_column("Std", justify="center", no_wrap=True)
        for item in self.items:
            table.add_row(
                item.nodeName,
                item.label,
                item.units,
                str(len(item.y)),
                str(round(np.sum(item.y), 3)),
                str(round(np.average(item.y), 3)),
                str(round(np.mean(item.y), 3)),
                str(round(np.var(item.y), 3)),
                str(round(np.std(item.y), 3))
            )
        console.print(table)
        return self

    def exportCDFExcel(self,
                       save_dir: str,
                       filename: str,
                       savePath: str,
                       xName: str = "Rate",
                       yName: str = "CDF"):
        """
        计算CDF（Cumulative Distribution Function）并导出
        :param xName:
        :param yName:
        :param savePath:
        :return:
        """
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        save_path = f"{save_dir}{os.sep}{filename}"
        longNames, units, labels = [], [], []
        ys, maxLen = [], 0

        for item in self.items:
            maxLen = max(maxLen, len(item.y))

        for item in self.items:
            y = np.array(np.pad(item.y, (0, maxLen - len(item.y)), 'constant', constant_values=(0, 0))).T
            x = np.sort(y)
            ys.append(x)
            longNames.append(xName)
            units.append(item.units)
            labels.append(item.label)
            y = 1. * np.arange(len(x)) / (len(x) - 1)
            ys.append(y)
            longNames.append(yName)
            units.append("")
            labels.append(item.label)

        contents = []
        # long name
        contents.append(longNames)

        # units
        contents.append(units)

        # comments
        contents.append(labels)

        dataMatrix = np.vstack(ys)

        # write datas
        for i in range(maxLen):
            contents.append(dataMatrix[:, i])

        df = pd.DataFrame(contents)
        df.to_excel(save_path, index=False, header=None)
        return self

    def exportExcel(self,
                    save_dir: str,
                    filename: str,
                    xName: str = "Times",
                    yName: str = "Rate",
                    xUnit: str = "Second"):
        """
        导出数据到Excel
        :param savePath:
        :param xName:
        :param yName:
        :param xUnit:
        :return:
        """
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        save_path = f"{save_dir}{os.sep}{filename}"
        contents = []
        nodeNames, units, labels, xResult = [], [], [], []
        for item in self.items:
            nodeNames.append(item.nodeName)
            units.append(item.units)
            labels.append(item.label)
            if len(item.x) > len(xResult):
                xResult = item.x
        # long name
        contents.append([xName] + [yName for _ in range(len(nodeNames))])

        # units
        contents.append([xUnit] + units)

        # comments
        contents.append([""] + labels)

        ys = [xResult]
        # padding y
        for item in self.items:
            ys.append(np.array(np.pad(item.y, (0, len(xResult) - len(item.y)), 'constant', constant_values=(0, 0))).T)

        dataMatrix = np.vstack(ys)

        # write datas
        for i in range(len(xResult)):
            contents.append(dataMatrix[:, i])

        df = pd.DataFrame(contents)
        df.to_excel(save_path, index=False, header=None)
        return self



"""
实现带宽利用率计算
"""
import math

from .throughput import ThroughputExtractBase, Throughput, NodeThroughput
from .common import GraphBase, ExportBase, NodeItem
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.legend import Legend
from matplotlib.patches import Patch
import matplotlib.ticker as mticker


class BandWidthUtilizationItem:
    """
    带宽利用率条目
    label => R1-R2
    nodeName => R1
    faceId => 271
    bandwidth => 100
    """

    def __init__(self, label: str, ticker: str, nodeName: str, faceId: int,
                 bandwidth: float, x: [float], y: [float]):
        self.label = label
        self.ticker = ticker
        self.nodeName = nodeName
        self.faceId = faceId
        self.bandwidth = bandwidth
        self.x = x
        self.y = y
        self.value = np.mean([item / bandwidth for item in y])


class BandwidthUtilizationExtractBase(ThroughputExtractBase):
    def __init__(self, throughput: Throughput):
        ThroughputExtractBase.__init__(self, throughput)
        self.bandwidth = 100


class BandwidthUtilizationGraph(BandwidthUtilizationExtractBase, GraphBase):
    """
    一个用于实现绘制带宽利用率图的类
    """

    def __init__(self, throughput: Throughput):
        GraphBase.__init__(self)
        BandwidthUtilizationExtractBase.__init__(self, throughput)
        self.totalWidth = 1
        self.figSize = (10, 7)  # 图的大小
        self.edgecolor = "black"  # 边缘颜色
        self.totalWidth = 0.8
        self.fig = None
        self.ax = None
        self.items: [BandWidthUtilizationItem] = []

    def xlim(self, *args, **kwargs):
        self.ax.set_xlim(*args, **kwargs)
        return self

    def ylim(self, *args, **kwargs):
        plt.ylim(*args, **kwargs)
        return self

    def xticks(self, ticks=None, labels=None, **kwargs):
        plt.xticks(ticks, labels, **kwargs)
        return self

    def title(self, label, fontdict=None, loc=None, pad=None, **kwargs):
        # self.ax.title(label, fontdict=fontdict, loc=loc, pad=pad, **kwargs)
        return self

    def legend(self, *args, **kwargs):
        handles, labels = self.ax.get_legend_handles_labels()
        self.ax.legend(handles, labels, *args, **kwargs)
        return self

    # def categoryLegend(self, facecolor='white', *args, loc='upper left', **kwargs):
    #     categories, patches = [], []
    #     for dtc in self.deliveryTimeCollections:
    #         categories.append(dtc.collectionName)
    #         patches.append(Patch(facecolor=facecolor, edgecolor=self.edgecolor, hatch=dtc.hatch))
    #     if "frameon" not in kwargs:
    #         kwargs["frameon"] = False
    #     leg = Legend(self.ax, patches, categories, *args, loc=loc, **kwargs)
    #     self.ax.add_artist(leg)
    #     return self

    def append(self, label: str, ticker: str, nodeName: str, faceId: int, bandwidth: float):
        """
        添加一个带宽利用率记录
        :param label:           dsccp
        :param ticker:          R1-R2
        :param nodeName:        R1
        :param faceId:          271
        :param bandwidth:       10
        :return:
        """
        node = self.getNode(nodeName)
        if not node:
            print("not exist node: ", node)
            return self
        x, y = self.getX(node, faceId), self.getY(node, faceId)
        self.items.append(BandWidthUtilizationItem(label, ticker, nodeName, faceId, bandwidth, x, y))
        return self

    def doDraw(self):
        """
        绘制带宽利用率对比图
        :return:
        """
        links, maxSize = dict(), 0
        colors, hatchs = dict(), dict()
        for item in self.items:
            if item.label not in links:
                links[item.label] = [item]
            else:
                links[item.label].append(item)
            if item.label not in colors:
                colors[item.label] = self.autoSelectColor()
                self.currentUsedColor[colors[item.label]] = item.label
            if item.label not in hatchs:
                hatchs[item.label] = self.autoSelectHatch()
            maxSize = max(maxSize, len(links[item.label]))

        xStart = np.arange(maxSize)
        total_width, n = self.totalWidth, len(links)
        width = total_width / n
        xStart = xStart - ((n - 1) * width / 2)

        self.fig, self.ax = plt.subplots(figsize=self.figSize)
        offsetX, isFirst = 0, True
        firstLabel = None
        for label in links:
            items = links[label]
            if firstLabel is None:
                firstLabel = label

            self.ax.bar(xStart + offsetX, [item.value for item in items], width=width, color=colors[label],
                        edgecolor=self.edgecolor,
                        label=items[0].label,
                        hatch=hatchs[label])
            offsetX += width

        consumers = [item.ticker for item in links[firstLabel]]
        print(consumers)
        xticklabels, nextTarget = [], 0
        ticks_loc = self.ax.get_xticks().tolist()
        for v in ticks_loc:
            if math.isclose(v, nextTarget, rel_tol=1e-5) and nextTarget < len(consumers):
                xticklabels.append(consumers[nextTarget])
                nextTarget += 1
            else:
                xticklabels.append("")
        print(ticks_loc)
        self.ax.xaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
        self.ax.set_xticklabels(xticklabels)
        return self

    def innerParse(self, inputFile: str):
        self.throughput = Throughput.parse(inputFile)
        return self

    @staticmethod
    def parse(inputFile: str):
        return BandwidthUtilizationGraph(Throughput.parse(inputFile))



from enum import Enum
from typing import Optional, List
from .common import GraphBase, NodeItem, ExportItem, ExportBase
import numpy as np
import pandas as pd


class DelayType(Enum):
    """
    延迟类型
    https://ndnsim.net/current/metric.html
    """
    # LastDelay means that DelayS and DelayUS represent delay between last Interest sent and Data packet received
    # LastDelay意味着DelayS和DelayUS代表最后发送的兴趣和接收的数据包之间的延迟
    LastDelay = 0
    # FullDelay means that DelayS and DelayUS represent delay between first Interest sent and Data packet received
    # (i.e., includes time of Interest retransmissions)
    # FullDelay是指DelayS和DelayUS代表发送的第一个感兴趣的数据包和接收的数据包之间的延迟
    FullDelay = 1


class DelayTarget(Enum):
    """
    延迟目标
    """
    DelayS = 0  # 按秒统计的延迟
    DelayMS = 1  # 按毫秒统计的延迟
    DelayUS = 2  # 按微秒统计的延迟

    @staticmethod
    def getUnit(delayTargetName: str) -> str:
        value = DelayTarget[delayTargetName].value
        if value == 0:
            return "Second"
        elif value == 1:
            return "Millisecond"
        elif value == 3:
            return "Microsecond"
        else:
            return "unknown"


class DelayItem:
    """
    一个用于描述 ndnsim 延迟采样结果的类
    https://ndnsim.net/current/metric.html#application-level-trace-helper
    """

    def __init__(self, Time: float, Node: str, AppId: int, SeqNo: int, Type: DelayType, DelayS: float,
                 DelayUS: float, RetxCount: int, HopCount: int):
        """
        :param Time:
        :param Node:
        :param AppId:
        :param SeqNo:
        :param Type:
        :param DelayS:
        :param DelayUS:
        :param RetxCount:
        :param HopCount:
        """
        self.Time = Time
        self.Node = Node
        self.AppId = AppId
        self.SeqNo = SeqNo
        self.Type = Type
        self.DelayS = DelayS
        self.DelayUS = DelayUS
        self.RetxCount = RetxCount
        self.HopCount = HopCount

    def getValueByDelayTarget(self, delayTarget: DelayTarget) -> float:
        if delayTarget == DelayTarget.DelayS:
            return self.DelayS
        elif delayTarget == DelayTarget.DelayMS:
            return self.DelayUS / 1000.0
        else:
            return self.DelayUS

    @staticmethod
    def parseLine(line: str):
        # 0.14809	C2	0	1	LastDelay	0.0610684	61068.4	1	3
        values = line.strip().split("\t")
        if len(values) < 9:
            return None
        return DelayItem(
            float(values[0].strip()),
            values[1].strip(),
            int(values[2].strip()),
            int(values[3].strip()),
            DelayType[values[4].strip()],
            float(values[5].strip()),
            float(values[6].strip()),
            int(values[7].strip()),
            int(values[8].strip())
        )


class NodeAppDelay:
    """
    用于描述某个节点上某个App的延迟 => 使用 AppId 来区分同一个节点上的不同App
    """

    def __init__(self, Node: str, AppId: int):
        self.Node = Node
        self.AppId = AppId
        self.typeMap = dict()  # 用于存储不同 Delay Type 的采样记录
        for delayType in DelayType:
            self.typeMap[delayType.name] = []

    def appendItem(self, item: DelayItem):
        """
        新增一条采样记录
        :param item:
        :return:
        """
        # 如果不是当前App的统计记录，忽略
        if item.Node != self.Node or item.AppId != self.AppId:
            return
        self.typeMap[item.Type.name].append(item)

    def _pre_sample(self, intervals: List[tuple], params: dict):
        delayType = params.get("delayType", DelayType.LastDelay)
        result = {
            'all': []
        }
        for interval in intervals:
            result[f"{interval[0]}-{interval[1]}"] = []
        for i in range(len(self.typeMap[DelayType.FullDelay.name])):
            fullItem, lastItem = self.typeMap[DelayType.FullDelay.name][i], \
                                 self.typeMap[DelayType.LastDelay.name][i]
            full, last = fullItem.getValueByDelayTarget(DelayTarget.DelayMS), \
                         lastItem.getValueByDelayTarget(DelayTarget.DelayMS)
            # 过滤重传的包
            if full - last > 10e-6:
                continue
            item = self.typeMap[delayType.name][i]
            result['all'].append(item)
            for interval in intervals:
                if len(interval) == 5 and (item.Time < interval[3] or item.Time > interval[4]):
                    continue
                if last < interval[0]:
                    break
                if last < interval[1]:
                    result[f"{interval[0]}-{interval[1]}"].append(item)
                    break
        return result

    def _sampling_interval(self, items: List[DelayItem], params: dict, current_res: dict) -> dict:
        lastCount, res, x_v, y_v, tmp = 0, {}, 0, 0, []
        delayType, samplingInterval, raw = \
            params.get("delayType", DelayType.LastDelay), params.get("samplingInterval", 2), \
            params.get("raw", -1)

        count = 0
        for item in items:
            last = item.getValueByDelayTarget(DelayTarget.DelayMS)
            if samplingInterval > 0:
                # 采样
                if raw == 0:
                    # 1. 采样区间内平均
                    if int(item.Time / samplingInterval) != lastCount:
                        lastCount = int(item.Time / samplingInterval)
                        x_v = lastCount * samplingInterval
                        y_v = np.average(tmp)
                        tmp = []
                        res[x_v] = y_v
                    else:
                        tmp.append(last)
                elif raw > 0:
                    # 2. 按样本数间隔采样
                    if count % samplingInterval == 0:
                        x_v = item.Time
                        y_v = last
                        res[x_v] = y_v
                else:
                    # 3. 按时间间隔采样
                    if int(item.Time / samplingInterval) != lastCount:
                        lastCount = int(item.Time / samplingInterval)
                        x_v = lastCount * samplingInterval
                        y_v = last
                        res[x_v] = y_v
            else:
                # 不采样，使用原始数据
                x_v = item.Time
                y_v = last
                res[x_v] = y_v
            count += 1
        return res

    def _sampling(self, params: dict, get_x=True):
        """
        采样
        :param params:
        :return:
        """
        lastCount, x, y = 0, [], []
        delayType, samplingInterval, raw, anomalyIntervalList = \
            params.get("delayType", DelayType.LastDelay), params.get("samplingInterval", 2), \
            params.get("raw", -1), params.get("anomalyIntervalList", [])

        # 处理分区
        if len(anomalyIntervalList) == 0:
            anomalyIntervalList = [(0, float("inf"))]
        elif anomalyIntervalList[0][0] != 0:
            anomalyIntervalList.insert(0, (0, anomalyIntervalList[0][0]))

        # 得到预采样的结果
        pre_sample_res = list(self._pre_sample(anomalyIntervalList[1:], params).values())

        # 1. 首先对全局进行采样
        result = self._sampling_interval(pre_sample_res[0], params, {})
        last_raw_count = len(pre_sample_res[0])
        last_sample_count = len(result)

        # 2. 接着对每个异常区间采样
        for i in range(1, len(pre_sample_res)):
            cur_raw_samples = pre_sample_res[i]
            params['raw'] = 1
            params['samplingInterval'] = int(last_raw_count / last_sample_count / 2)
            if params['samplingInterval'] > len(cur_raw_samples) / 10:
                params['samplingInterval'] = int(len(cur_raw_samples) / 10)
            if len(anomalyIntervalList[i]) >= 3:
                params['samplingInterval'] = anomalyIntervalList[i][2]
            res = self._sampling_interval(cur_raw_samples, params, result)
            # if len(res) > 0:
            #     last_raw_count = len(cur_raw_samples)
            #     last_sample_count = len(res)
            result.update(res)
        params['raw'] = raw
        params['samplingInterval'] = samplingInterval

        # 3. 按时间排序
        keys = list(result.keys())
        keys.sort()
        x = keys
        if not get_x:
            for k in keys:
                y.append(result[k])

        return x, y

    def getX(self, params: dict):
        """
        获取采样时间列表
        :param delayType:
        :return:
        """
        x, _ = self._sampling(params, True)
        return x

    def getY(self, params: dict):
        """
        获取延迟采样列表
        :param delayType:           延迟类型
        :param delayTarget:         延迟目标
        :param samplingInterval:    采样间隔
        :return:
        """
        _, y = self._sampling(params, False)
        return y


class NodeDelay:
    """
    一个用于描述同一个节点上各个App延迟的类
    """

    def __init__(self, Node: str):
        self.Node = Node
        self.appMap = dict()

    def appendItem(self, item: DelayItem):
        """
        新增一条采样记录
        :param item:
        :return:
        """
        # 如果不是当前节点的采样记录，则忽略
        if item.Node != self.Node:
            return
        if item.AppId not in self.appMap:
            self.appMap[item.AppId] = NodeAppDelay(item.Node, item.AppId)
        self.appMap[item.AppId].appendItem(item)

    def getX(self, AppId: int, params: dict):
        """
        获取某个App的采样时间列表
        :param AppId:
        :param delayType:
        :param samplingInterval:
        :return:
        """
        if AppId not in self.appMap:
            return []
        return self.appMap[AppId].getX(params)

    def getY(self, AppId: int, params: dict):
        """
        获取某个App的延迟采样列表
        :param AppId:
        :param delayType:
        :param delayTarget:
        :param samplingInterval:
        :return:
        """
        if AppId not in self.appMap:
            return []
        return self.appMap[AppId].getY(params)


class Delay:
    """
    一个用于解析 ndnsim 延迟采样结果的类
    """

    def __init__(self):
        self.nodeMap = dict()

    def getByNode(self, node: str) -> Optional[NodeDelay]:
        if node not in self.nodeMap:
            return None
        return self.nodeMap[node]

    @staticmethod
    def parse(inputFile: str):
        """
        传入一个 ndnsim 采集的延迟采样结果文件，从中解析出延迟信息
        :param inputFile:
        :return:
        """
        delay = Delay()
        with open(inputFile, "r") as f:
            # 首先忽略表头
            # Time	Node	AppId	SeqNo	Type	DelayS	DelayUS	RetxCount	HopCount
            f.readline()

            # 从第一行开始解析
            line = f.readline()
            while line:
                if line.strip() != "":
                    item = DelayItem.parseLine(line)
                    if item:
                        if item.Node not in delay.nodeMap:
                            delay.nodeMap[item.Node] = NodeDelay(item.Node)
                        delay.nodeMap[item.Node].appendItem(item)
                line = f.readline()
        return delay


class DelayExtractBase:
    def __init__(self, delay: Delay):
        self.samplingInterval = 1.0
        self.delayType = DelayType.LastDelay
        self.delayTarget = DelayTarget.DelayMS
        self.delay = delay
        self.raw = 0
        # 异常区间
        # 1. 在进行采样时，异常区间列表中的每一个区间单独采样；
        self.anomalyIntervalList: List[tuple] = []

    def setAnomalyInterval(self, intervals: List[tuple]):
        """
        设置异常采样区间
        :param intervals:
        :return:
        """
        self.anomalyIntervalList = intervals
        return self

    def setSamplingInterval(self, samplingInterval: float):
        """
        设置采样间隔 => 如果 samplingInterval = 1.0 表示绘图时，每秒采样一次
        :param samplingInterval:
        :return:
        """
        self.samplingInterval = samplingInterval
        return self

    def setRaw(self, raw: int):
        """
        raw < 0 => 按时间间隔采样
        raw > 0 => 按样本数间隔采样
        raw == 0 => 平均采样
        :param raw:
        :return:
        """
        self.raw = raw
        return self

    def setDelayType(self, delayType: DelayType):
        """
        设置延迟类型
        :param delayType:
        :return:
        """
        self.delayType = delayType
        return self

    def setDelayTarget(self, delayTarget: DelayTarget):
        """
        设置延迟目标
        :param delayTarget:
        :return:
        """
        self.delayTarget = delayTarget
        return self

    def getNode(self, nodeName):
        return self.delay.getByNode(nodeName)

    def getX(self, node: NodeDelay, nodeId):
        return node.getX(nodeId, self.__dict__)

    def getY(self, node: NodeDelay, nodeId):
        return node.getY(nodeId, self.__dict__)

    def getUnit(self):
        return DelayTarget.getUnit(self.delayTarget.name)


class DelayGraph(GraphBase, DelayExtractBase, ExportBase):
    """
    一个用于实现绘制延迟图的类
    """

    def __init__(self, delay: Delay):
        GraphBase.__init__(self)
        DelayExtractBase.__init__(self, delay)
        ExportBase.__init__(self, self.getNode, self.getX, self.getY, self.getUnit)

    def plot(self, nodeName: str, appId: int, *args,
             color: str = "&&",
             linewidth: float = 2, linestyle: str = "dotted", marker: str = "&&",
             markerfacecolor: str = "none", markersize: float = 6,
             **kwargs):
        node = self.delay.getByNode(nodeName)
        if not node:
            print("not exist node: ", node)
            return self
        if "label" not in kwargs:
            kwargs["label"] = node.Node
        x, y = node.getX(appId, self.__dict__), \
               node.getY(appId, self.__dict__)
        self.select(nodeName, appId, kwargs["label"])
        super().innerPlot(x, y, *args,
                          color=color,
                          linewidth=linewidth,
                          linestyle=linestyle,
                          marker=marker,
                          markerfacecolor=markerfacecolor,
                          markersize=markersize,
                          **kwargs)
        return self

    def plotSum(self, nodeList: [NodeItem], *args,
                color: str = "&&",
                linewidth: float = 2, linestyle: str = "dotted", marker: str = "&&",
                markerfacecolor: str = "none", markersize: float = 6,
                **kwargs):
        label = None
        if "label" in kwargs:
            label = kwargs["label"]
        self.selectSum(nodeList, label=label)
        super().innerPlotSum(nodeList, self.getNode, self.getX, self.getY, *args,
                             color=color,
                             linewidth=linewidth,
                             linestyle=linestyle,
                             marker=marker,
                             markerfacecolor=markerfacecolor,
                             markersize=markersize,
                             **kwargs)
        return self

    def plotAvg(self, nodeList: [NodeItem], *args,
                color: str = "&&",
                linewidth: float = 2, linestyle: str = "dotted", marker: str = "&&",
                markerfacecolor: str = "none", markersize: float = 6,
                **kwargs):
        label = None
        if "label" in kwargs:
            label = kwargs["label"]
        self.selectAvg(nodeList, label=label)
        self.innerPlotAvg(nodeList, self.getNode, self.getX, self.getY, *args,
                          color=color,
                          linewidth=linewidth,
                          linestyle=linestyle,
                          marker=marker,
                          markerfacecolor=markerfacecolor,
                          markersize=markersize,
                          **kwargs)
        return self

    def exportExcel(self, savePath: str,
                    xName: str = "Times",
                    yName: str = "Delay",
                    xUnit: str = "Second"):
        """
        导出数据到Excel
        :param savePath:
        :param xName:
        :param yName:
        :param xUnit:
        :return:
        """
        dfResult = None
        for item in self.items:
            contents = [[xName, yName], [xUnit, item.units], ["", item.label]]
            dataMatrix = np.vstack([np.array(item.x).T, np.array(item.y).T])
            for i in range(len(item.x)):
                contents.append(dataMatrix[:, i])
            df = pd.DataFrame(contents)
            if dfResult is None:
                dfResult = df
            else:
                dfResult = pd.concat([dfResult, df], axis=1)
        dfResult.to_excel(savePath, index=False, header=None)
        return self

    def exportGroupExcel(self, savePath: str):
        dfResult = None
        for item in self.items:
            nextCount = int(item.x[0] / self.samplingInterval) * self.samplingInterval + self.samplingInterval
            # 填充没有的时间
            for i in range(int(item.x[0] / self.samplingInterval)):
                res = [[(i + 1) * self.samplingInterval], [item.label], []]
                df = pd.DataFrame(res)
                if dfResult is None:
                    dfResult = df
                else:
                    dfResult = pd.concat([dfResult, df], axis=1)
            tmp = [[nextCount], [item.label], []]
            for i in range(len(item.x)):
                if item.x[i] < nextCount:
                    tmp.append([item.y[i]])
                else:
                    df = pd.DataFrame(tmp)
                    if dfResult is None:
                        dfResult = df
                    else:
                        dfResult = pd.concat([dfResult, df], axis=1)
                    nextCount += self.samplingInterval
                    tmp = [[nextCount], [item.label], [], [item.y[i]]]
            if len(tmp) > 10:
                dfResult = pd.concat([dfResult, pd.DataFrame(tmp)], axis=1)
        dfResult.to_excel(savePath, index=False, header=None)
        return self

    def exportGroupCategoriesExcel(self, savePath: str):
        dfResult = None
        tmp = [['Time', 'Consumer', 'Data'], ['', '', ''], ['', '', '']]
        for item in self.items:
            nextCount = int(item.x[0] / self.samplingInterval) * self.samplingInterval + self.samplingInterval
            # 填充没有的时间
            for i in range(int(item.x[0] / self.samplingInterval)):
                res = [[(i + 1) * self.samplingInterval], [item.label], []]
                tmp.append([self.samplingInterval * (i + 1), item.label, 0])

            for i in range(len(item.x)):
                tmp.append([nextCount, item.label, item.y[i]])
                if item.x[i] < nextCount:
                    continue
                else:
                    nextCount += self.samplingInterval

        pd.DataFrame(tmp).to_excel(savePath, index=False, header=None)
        return self

    @staticmethod
    def parse(inputFile: str):
        return DelayGraph(Delay.parse(inputFile))
